RESEARCH_SESSION_CONFIG = [
    {
        'construction': {
            'model': None,
            'optimizer': None,
            'criterion': None,
            'watch_test': None,
            'research_scheme': None
        },
        'session': {
            'train_dataloader': None,
            'test_dataloader': None,
            'dataset_name': None,
            'epochs': None,
            'iteration_modulo': None
        }

    }
]



import matplotlib
matplotlib.use("TkAgg")

import matplotlib.pyplot as plt
plt.ion()

import numpy as np
import time


__all__ = ['Visualizer']


class Func:
    def __init__(self, axes, color='b', name=None):
        self.x = np.array([])
        self.y = np.array([])
        self.color = color
        self.axes = axes
        self.curve = self.axes.plot(self.x, self.y, self.color, label=name)[0]
        self.axes.legend(loc='best')

    def update(self, new_x, new_y):
        self.x = np.append(self.x, new_x)
        self.y = np.append(self.y, new_y)
        self.curve.set_xdata(self.x)
        self.curve.set_ydata(self.y)
        self._rescale()

    def _rescale(self):
        self.axes.relim()
        self.axes.autoscale_view()


class Visualizer:
    def __init__(self, curves_descriptors: list, x_label=None, y_label=None):
        self.fig = plt.figure()
        self.axes = self.fig.add_subplot(1, 1, 1)
        self.axes.set_xlabel(x_label)
        self.axes.set_ylabel(y_label)
        self.axes.autoscale(enable=True, axis='both')
        self.funcs = {d['name']: Func(
            self.axes, d['color'], d['name']) for d in curves_descriptors}

    def update(self, new_values: dict):
        for func_name in new_values:
            func = self.funcs[func_name]
            values = new_values[func_name]
            func.update(values['x'], values['y'])

    def redraw(self):
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()


if __name__ == '__main__':
    funcs_descriptors = [
        {'name': 'train', 'color': 'r'},
        {'name': 'test', 'color': 'b'},
    ]
    visualizer = Visualizer(funcs_descriptors, 'x_label', 'y_label')

    def rand(): return np.random.uniform(0, 10)

    for i in range(10):
        visualizer.update({
            'train': {'x': rand(), 'y': rand()},
            'test': {'x': rand(), 'y': rand()},
        })
        visualizer.redraw()
        time.sleep(1)



"""
    This module contains tools for
    models examination and comparison.

    Using a research scheme you can
    accomplish research of a given
    model in customizable manner.

    Using report scheme you can include
    specific scheme to the report of the
    research.

    * It's kinda network scheme in keras.

"""

import os
import torch
import matplotlib.pyplot as plt
import importlib
from .path_generation import ModelPathGenerator
from .decorators import unfinished_class


# TODO: use kwargs from Research as SUPER kwargs. So they have more priority with regard of items' kwargs


class Research:
    def __init__(self,
                 research_path: str,
                 research_scheme: list):
        self._research_path = research_path
        if not os.path.exists(self._research_path):
            os.makedirs(self._research_path)
        self._research_scheme = research_scheme

    def start_research_session(self):
        raise NotImplementedError(
            "start_research_session(...)"
            " should be implemented."
        )


class ModelResearch(Research):
    def __init__(self,
                 research_path: str,
                 research_scheme: list,
                 model,
                 optimizer,
                 criterion,
                 watch_test=None):
        Research.__init__(self, research_path, research_scheme)
        self._model = model
        self._optimizer = optimizer
        self._criterion = criterion
        self._use_cuda = next(model.parameters()).is_cuda
        self._watch_test = watch_test

        # it's used by research items
        if watch_test is not None:
            for item in self._research_scheme:
                item._kwargs['watch_test'] = watch_test

        # save initial model
        internal_path = os.path.join(self._research_path, 'model_init.pth')
        torch.save(self._model.state_dict(), internal_path)

    def start_research_session(self, train_dataloader, test_dataloader=None, epochs=1, iteration_modulo=None, **kwargs):
        """

        Arguments:
            train_dataloader {} -- 

        Keyword Arguments:
            test_dataloader {} --  (default: {None})
            epochs {} --  (default: {1})
            iteration_modulo {} -- per give count of iterations research items will be called
        """

        # update iteration modulo if needed
        if iteration_modulo is not None:
            for item in self._research_scheme:
                item._kwargs['iteration_modulo'] = 1
        else:
            iteration_modulo = 1

        # clear report files
        for item in self._research_scheme:
            internal_path = item._kwargs.get('internal_path', 'report.txt')
            report_path = os.path.join(self._research_path, internal_path)
            if os.path.exists(report_path):
                open(report_path, 'w').close()

        # This dict will be
        # sent to the research
        # items in the research scheme
        network_output = {
            'iteration': None,
            'epoch': None,
            'accuracy_train': None,
            'accuracy_test': None,
            'loss_train': None,
            'loss_test': None,
            'model': self._model,
            'optimizer': self._optimizer,
            'criterion': self._criterion,
            'input': None,
            'output': None,
            'expected_output': None,
            'iterations_per_epoch': min(len(train_dataloader), len(test_dataloader))
        }

        # TODO: here is too many `if self._watch_test:` conditions

        # If only train is needed
        if self._watch_test:
            dataloaders = zip(train_dataloader, test_dataloader)
        else:
            dataloaders = train_dataloader
            network_output['iterations_per_epoch'] = len(train_dataloader)

        # Main cycle
        running_loss_train = 0.0
        if self._watch_test:
            running_loss_test = 0.0
        # TODO: test dataloader has fewer number of samples than train dataloader

        for epoch in range(epochs):
            for i, data in enumerate(dataloaders):

                if self._watch_test:
                    ((X_train, y_train), (X_test, y_test)) = data
                else:
                    (X_train, y_train) = data

                network_output['iteration'] = i
                network_output['epoch'] = epoch
                network_output['input'] = X_train
                network_output['expected_output'] = y_train
                # if using cuda
                if self._use_cuda:
                    X_train = X_train.cuda()
                    y_train = y_train.cuda()
                    if self._watch_test:
                        X_test = X_test.cuda()
                        y_test = y_test.cuda()

                # zero gradients of the parameters
                self._optimizer.zero_grad()

                # calculate the loss
                y_pred = self._model(X_train)
                if self._watch_test:
                    y_pred_test = self._model(X_test)

                if self._use_cuda:
                    y_pred = y_pred.cuda()
                    if self._watch_test:
                        y_pred_test = y_pred_test.cuda()

                network_output['output'] = y_pred

                loss_train = self._criterion(y_pred, y_train)
                if self._watch_test:
                    loss_test = self._criterion(y_pred_test, y_test)

                running_loss_train += loss_train
                if self._watch_test:
                    running_loss_test += loss_test

                # backward propagation
                loss_train.backward()

                # parameters optimization
                self._optimizer.step()

                # absolute iteration
                absolute_iteration = network_output['iterations_per_epoch'] * epoch + i

                if absolute_iteration % iteration_modulo == 0:
                    network_output['loss_train'] = running_loss_train / \
                        iteration_modulo
                    if self._watch_test:
                        network_output['loss_test'] = running_loss_test / \
                            iteration_modulo

                    # write to the report file
                    for item in self._research_scheme:
                        item.print(self._research_path, **network_output)

                    running_loss_train = 0.0

                    if self._watch_test:
                        running_loss_test = 0.0

        # save the final model
        model_path = os.path.join(
            self._research_path,
            'model.pth'
        )
        torch.save(self._model.state_dict(), model_path)


@unfinished_class
class ResearchSession:
    def __init__(self, config_module: str, root_path: str):
        self._config_module = config_module
        self._root_path = root_path

    def start_research_session(self):
        # here we process a banach of
        # models written in config files
        research_list = self.__parse_config_file()
        for (research, kwargs) in research_list:
            research.start_research_session(**kwargs)

    def __parse_config_file(self):
        # importlib.reload(self._config_module)
        research_list = []
        for session_description in self._config_module.RESEARCH_SESSION_CONFIG:
            session_kwargs = session_description['session']
            construct_kwargs = session_description['construction']

            # Generate research path name
            path_generator = ModelPathGenerator(
                root_path=self._root_path
            )
            research_path = path_generator.generate_path(
                model=construct_kwargs['model'],
                optimizer=construct_kwargs['optimizer'],
                criterion=construct_kwargs['criterion'],
                epochs=session_kwargs['epochs'],
                dataset_name=session_kwargs.get('dataset_name', None),
                lr=session_kwargs.get('lr_in_path', True),
                makedirs=True
            )

            research = ModelResearch(
                research_path=research_path, **construct_kwargs)
            research_list.append((research, session_kwargs))
        return research_list


@unfinished_class
class ModelsComparison(Research):
    # TODO: Fix everything
    def __init__(self,
                 research_path: str,
                 research_scheme: list,
                 models_dir):
        Research.__init__(self, research_scheme)

        self._models_dir = models_dir

    def start_research_session(self):
        pass


if __name__ == "__main__":
    ### Example ###

    # modules for testing
    import torchvision.models as models
    import torchvision.transforms as transforms
    import torchvision

    # items
    from .research_items import ModelConfigurationItem, CurrentIterationItem, LossPrintItem, LossVisualizationItem

    # net specific stuff
    net = torchvision.models.MobileNetV2(num_classes=100).cuda()
    optimizer = torch.optim.SGD(params=net.parameters(), lr=0.01)
    criterion = torch.nn.CrossEntropyLoss()

    # load the data
    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(
            mean=(0.485, 0.456, 0.406),
            std=(0.229, 0.224, 0.225)
        )
    ])
    trainset = torchvision.datasets.CIFAR100(root='./data/datasets', train=True,
                                             download=True, transform=transform)
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=32,
                                              shuffle=True, num_workers=2)
    testset = torchvision.datasets.CIFAR100(root='./data/datasets', train=False,
                                            download=True, transform=transform)
    testloader = torch.utils.data.DataLoader(testset, batch_size=32,
                                             shuffle=False, num_workers=2)

    # MAGIC
    research = ModelResearch(
        research_path='trash',  # path to the folder with results
        research_scheme=[  # what to include in report
            ModelConfigurationItem(),
            CurrentIterationItem(),
            LossPrintItem(),
            LossVisualizationItem()
        ],
        model=net,
        optimizer=optimizer,
        criterion=criterion
    )

    research.start_research_session(trainloader, testloader, epochs=10)



# Different types of research
from .research import Research
from .research import ModelResearch
from .research import ResearchSession

# Research items
from .research_items import ResearchItem
from .research_items import AccuracyPrintItem
from .research_items import LossPrintItem
from .research_items import LossVisualizationItem
from .research_items import CurrentIterationItem
from .research_items import ModelConfigurationItem



"""

Note: I need some tools to generate
paths. e.g. to generate paths for 
researches for differnet models.
"""

import os


class PathGenerator:
    def __init__(self, root_path):
        self._root_path = root_path

    def generate_path(self, makedirs: bool = False):
        raise NotImplementedError("PathGenerator cannot be instanciated")


class DefaultPathGenerator(PathGenerator):
    def generate_path(self, internal_path: str, makedirs: bool = False):
        path = os.path.join(
            self._root_path,
            internal_path
        )
        if makedirs and not os.path.exists(path):
            os.makedirs(path)
        return path


class ModelPathGenerator(PathGenerator):
    """Generates path to folder for model 
    log files.
    """

    def generate_path(self,
                      model,
                      optimizer=None,
                      criterion=None,
                      epochs: int = None,
                      dataset_name: str = None,
                      lr: bool = False,
                      makedirs=False):
        path = model.__class__.__name__
        if optimizer is not None:
            path += '_' + optimizer.__class__.__name__
            if lr is True:
                # part of learning rate after '.'
                path += str(optimizer.defaults['lr']).split('.')[1]
        if criterion is not None:
            path += '_' + criterion.__class__.__name__
        if epochs is not None:
            path += '_' + str(epochs) + 'ep'
        if dataset_name is not None:
            path += '_' + dataset_name

        path = os.path.join(self._root_path, path)

        if makedirs and not os.path.exists(path):
            os.makedirs(path)
        return path


if __name__ == "__main__":
    import torch
    import torchvision
    import torchvision.transforms as transforms

    net = torchvision.models.MobileNetV2(num_classes=100).cuda()

    optimizer = torch.optim.SGD(
        params=net.parameters(),
        lr=0.001,
        momentum=0.9
    )
    criterion = torch.nn.CrossEntropyLoss()

    gen = ModelPathGenerator('!research')
    path = gen.generate_path(net, optimizer, criterion, 10, 'MNIST')
    print(path)




def unfinished_function(func):
    def result(*args, **kwargs):
        print('Function {} is unfinished'.format(func.__name__) +
              'and not enough tested. '
              'Be careful with it.')
        func(*args, **kwargs)
    return result


def unfinished_class(cls):
    def new(cls, *args, **kwargs):
        result = object.__new__(cls)
        cls.__init__(result, *args, **kwargs)
        print('Class {} is unfinished'.format(cls.__name__) +
              'and not enough tested. '
              'Be careful with it.')
        return result
    cls.__new__ = new
    return cls


if __name__ == "__main__":
    @unfinished_class
    class X:
        def __init__(self, x):
            self.x = x

    print(X(10).x)



from distutils.core import setup
setup(
    name='pytorchresearch',         # How you named your package folder (MyLib)
    packages=['pytorchresearch'],   # Chose the same as "name"
    # Start with a small number and increase it with every change you make
    version='1.0.3-alpha',
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    license='MIT',
    # Give a short description about your library
    description='Tools to automate research with PyTorch framework',
    author='Ilya Vouk',                   # Type in your name
    author_email='ilya.vouk@gmail.com',      # Type in your E-Mail
    # Provide either the link to your github or to your website
    url='https://github.com/VoIlAlex/pytorchresearch',
    # I explain this later on
    download_url='https://github.com/VoIlAlex/pytorchresearch/archive/v1.0.0-alpha.tar.gz',
    # Keywords that define your package best
    keywords=['framework', 'automation', 'research', 'science',
              'machine learning', 'deep learning', 'pytorch'],
    install_requires=[            # I get to this in a second
        'numpy',
        'matplotlib',
        'torch'
    ],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        # Specify which python versions that you want to support
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import os
from .visualizer import Visualizer
import matplotlib.pyplot as plt
from .decorators import unfinished_class, unfinished_function


class ResearchItem:
    """Base class for all the research
    items.
    """

    def print(self, research_path, *args, **kwargs):
        raise NotImplementedError(
            'print(...) method should be implemented '
            'by all the childs of ResearchItem class'
        )

    def final_action(self, research_path, *args, **kwargs):
        # TODO: Add final action support for items
        None


class CurrentIterationItem(ResearchItem):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def print(self, research_path, *args, **kwargs):

        #
        # extract arguments
        #

        # user provided arguments
        print_end = self._kwargs.get('print_end', '\n')
        internal_path = self._kwargs.get('internal_path', 'report.txt')
        iteration_modulo = self._kwargs.get('iteration_modulo', 1)
        use_absolute_iteration = self._kwargs.get('absolute_iteration', False)
        iterations_per_epoch = kwargs.get('iterations_per_epoch', None)

        # framework provided arguments
        epoch = kwargs.get('epoch', -1)
        iteration = kwargs.get('iteration', -1)

        # absolute iterations count
        absolute_iteration = iterations_per_epoch * epoch + iteration

        # get out if necessary
        if iteration % iteration_modulo != 0:
            return

        # path to file with the report
        file_path = os.path.join(
            research_path,
            internal_path
        )

        # write item to the file
        if use_absolute_iteration:
            with open(file_path, 'a') as f:
                print('Iteration: {:10d}'.format(absolute_iteration),
                      file=f,
                      end=print_end)
        else:
            with open(file_path, 'a') as f:
                print('Iteration: [{:6d},{:6d}]'.format(
                    epoch, iteration),
                    file=f,
                    end=print_end)


class AccuracyPrintItem(ResearchItem):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def print(self, research_path, *args, **kwargs):

        #
        # extract arguments
        #

        # user provided arguments
        print_end = self._kwargs.get('print_end', '\n')
        internal_path = self._kwargs.get('internal_path', 'report.txt')
        iteration_modulo = self._kwargs.get('iteration_modulo', 1)
        watch_test = self._kwargs.get('watch_test', False)

        # framework provided arguments
        iteration = kwargs.get('iteration', -1)
        accuracy_train = kwargs.get('accuracy_train', -1)
        accuracy_test = kwargs.get('accuracy_test', -1)

        # TODO: use absolute iteration
        # get out if necessary
        if iteration % iteration_modulo != 0:
            return

        # path to file with the report
        file_path = os.path.join(
            research_path,
            internal_path
        )

        # write item to the file
        with open(file_path, 'a') as f:
            if watch_test:
                print('Accuracy: [{:6.2f},{:6.2f}] '.format(accuracy_train, accuracy_test),
                      file=f,
                      end=print_end)
            else:
                print('Accuracy: {:6.2f}'.format(accuracy_train),
                      file=f,
                      end=print_end)


class LossPrintItem(ResearchItem):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def print(self, research_path, *args, **kwargs):

        #
        # extract arguments
        #

        # user provided arguments
        print_end = self._kwargs.get('print_end', '\n')
        internal_path = self._kwargs.get('internal_path', 'report.txt')
        iteration_modulo = self._kwargs.get('iteration_modulo', 1)
        watch_test = self._kwargs.get('watch_test', False)

        # framework provided arguments
        iteration = kwargs.get('iteration', -1)
        loss_train = kwargs.get('loss_train', None)
        loss_test = kwargs.get('loss_test', None)

        # get out if necessary
        if iteration % iteration_modulo != 0:
            return

        # path to file with the report
        file_path = os.path.join(
            research_path,
            internal_path
        )

        # write item to the file
        with open(file_path, 'a') as f:
            if watch_test:
                print('Loss: [{:6.2f},{:6.2f}] '.format(loss_train, loss_test),
                      file=f,
                      end=print_end)
            else:
                print('Loss: {:6.2f}'.format(loss_train),
                      file=f,
                      end=print_end)


class LossVisualizationItem(ResearchItem):
    def __init__(self, watch_test: bool = False, img_name: str = 'loss_rate.png', *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        if watch_test is True:
            func_descriptors = [
                {
                    'name': 'train',
                    'color': 'r'
                },
                {
                    'name': 'test',
                    'color': 'b'
                }
            ]
        else:
            func_descriptors = [
                {
                    'name': 'train',
                    'color': 'r'
                }
            ]
        self._watch_test = watch_test
        self._visualizer = Visualizer(func_descriptors, 'Iterations', 'Loss')
        self._research_path = 'loss_rate.png'
        self._img_name = img_name

    def print(self, research_path, *args, **kwargs):
        self._research_path = research_path
        # user provided arguments
        print_end = self._kwargs.get('print_end', '\n')
        internal_path = self._kwargs.get(
            'internal_path', 'report.txt')
        iteration_modulo = self._kwargs.get('iteration_modulo', 1)

        # framework provided arguments
        iteration = kwargs.get('iteration', -1)
        iterations_per_epoch = kwargs.get('iterations_per_epoch', None)
        epoch = kwargs.get('epoch', -1)
        loss_train = kwargs.get('loss_train', None)
        loss_test = kwargs.get('loss_test', None)

        # convert the loss
        # to the right format
        loss_train = loss_train.cpu().detach().numpy()
        if self._watch_test:
            loss_test = loss_test.cpu().detach().numpy()

        # absolute iteration
        x_pos = iterations_per_epoch * epoch + iteration

        if iteration % iteration_modulo == 0:
            if self._watch_test is True:
                self._visualizer.update({
                    'train': {
                        'x': x_pos,
                        'y': loss_train
                    },
                    'test': {
                        'x': x_pos,
                        'y': loss_test
                    }
                })
            else:
                self._visualizer.update({
                    'train': {
                        'x': x_pos,
                        'y': loss_train
                    }
                })
        self._visualizer.redraw()

    # def __del__(self):
    #     # fig_file_name = os.path.join(
    #     #     self._research_path,
    #     #     self._img_name
    #     # )
    #     # plt.savefig(fig_file_name)
    #     pass


class ModelConfigurationItem(ResearchItem):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def print(self, research_path, *args, **kwargs):
        #
        # extract arguments
        #

        # user provided arguments
        print_end = self._kwargs.get('print_end', '\n')
        internal_path = self._kwargs.get('internal_path', 'report.txt')

        # framework provided arguments
        epoch = kwargs.get('epoch', 0)
        iteration = kwargs.get('iteration', 0)

        # it should be printed only once
        if epoch != 0 or iteration != 0:
            return

        model = kwargs.get('model', None)
        optimizer = kwargs.get('optimizer', None)
        criterion = kwargs.get('criterion', None)

        # path to file with the report
        file_path = os.path.join(
            research_path,
            internal_path
        )

        # write item to the file
        with open(file_path, 'a') as f:
            print(model, file=f)
            print(optimizer, file=f)
            print(criterion, file=f, end=print_end)
